@@ -17,9 +17,12 @@
  */
 
 #include "gtd-application.h"
-#include "gtd-list-view.h"
+#include "gtd-task-list-view.h"
 #include "gtd-manager.h"
+#include "gtd-notification.h"
+#include "gtd-notification-widget.h"
 #include "gtd-storage-dialog.h"
+#include "gtd-task.h"
 #include "gtd-task-list.h"
 #include "gtd-task-list-item.h"
 #include "gtd-window.h"
@@ -34,25 +37,22 @@ typedef struct
   GtkFlowBox                    *lists_flowbox;
   GtkStack                      *main_stack;
   GtkWidget                     *new_list_popover;
-  GtkButton                     *notification_action_button;
-  GtkLabel                      *notification_label;
-  GtkRevealer                   *notification_revealer;
-  GtkSpinner                    *notification_spinner;
-  GtdListView                   *scheduled_list_view;
+  GtdNotificationWidget         *notification_widget;
+  GtdTaskListView               *scheduled_list_view;
+  GtkSearchBar                  *search_bar;
+  GtkToggleButton               *search_button;
+  GtkSearchEntry                *search_entry;
+  GtkStack                      *stack;
   GtkStackSwitcher              *stack_switcher;
   GtdStorageDialog              *storage_dialog;
-  GtdListView                   *today_list_view;
-  GtdListView                   *list_view;
+  GtdTaskListView               *today_list_view;
+  GtdTaskListView               *list_view;
 
   /* mode */
   GtdWindowMode                  mode;
 
-  /*
-   * A queue of the next operations.
-   */
-  GQueue                        *notification_queue;
-  gint                           notification_delay_id;
-  gboolean                       consuming_notifications;
+  /* loading notification */
+  GtdNotification               *loading_notification;
 
   GtdManager                    *manager;
 } GtdWindowPrivate;
@@ -65,36 +65,10 @@ struct _GtdWindow
   GtdWindowPrivate     *priv;
 };
 
-typedef struct
-{
-  GtdWindow   *window;
-
-  gchar       *id;
-  gchar       *text;
-  gchar       *label;
-
-  gint         delay;
-  gboolean     show_spinner;
-
-  GSourceFunc  primary_action;
-  GSourceFunc  secondary_action;
-
-  gpointer     data;
-} NotificationData;
-
-#define LOADING_LISTS_NOTIFICATION_ID            "loading-lists-id"
-
 static void          gtd_window__change_storage_action           (GSimpleAction         *simple,
                                                                   GVariant              *parameter,
                                                                   gpointer               user_data);
 
-static gboolean      gtd_window__execute_notification_data       (NotificationData      *data);
-
-
-static void          gtd_window_consume_notification             (GtdWindow             *window);
-
-void                 notification_data_free                      (NotificationData      *data);
-
 G_DEFINE_TYPE_WITH_PRIVATE (GtdWindow, gtd_window, GTK_TYPE_APPLICATION_WINDOW)
 
 static const GActionEntry gtd_window_entries[] = {
@@ -107,15 +81,74 @@ enum {
   LAST_PROP
 };
 
-void
-notification_data_free (NotificationData *data)
+static void
+gtd_window__stack_visible_child_cb (GtdWindow *window)
 {
-  g_free (data->id);
-  g_free (data->text);
-  g_free (data->label);
-  g_free (data);
+  GtdWindowPrivate *priv;
+
+  priv = window->priv;
+
+  gtk_widget_set_visible (
+          GTK_WIDGET (priv->search_button),
+          g_strcmp0 (gtk_stack_get_visible_child_name (priv->main_stack), "overview") == 0 &&
+          g_strcmp0 (gtk_stack_get_visible_child_name (priv->stack), "lists") == 0);
+
 }
 
+static void
+gtd_window_update_list_counters (GtdTaskList *list,
+                                 GtdTask     *task,
+                                 GtdWindow   *window)
+{
+  GtdWindowPrivate *priv;
+  GtkWidget *container_child;
+  gboolean is_today;
+  GList *tasks;
+  GList *l;
+  gchar *new_title;
+  gint counter;
+
+  g_return_if_fail (GTD_IS_WINDOW (window));
+  g_return_if_fail (GTD_IS_TASK_LIST (list));
+
+  priv = window->priv;
+
+  /* Count the number of incomplete tasks */
+  counter = 0;
+  tasks = gtd_task_list_get_tasks (list);
+
+  for (l = tasks; l != NULL; l = l->next)
+    {
+      if (!gtd_task_get_complete (l->data))
+        counter++;
+    }
+
+  /* Update the list title */
+  is_today = list == gtd_manager_get_today_list (priv->manager);
+
+  container_child = is_today ? GTK_WIDGET (priv->today_list_view) : GTK_WIDGET (priv->scheduled_list_view);
+
+  if (counter == 0)
+    {
+      new_title = g_strdup_printf ("%s", is_today ? _("Today") : _("Scheduled"));
+    }
+  else
+    {
+      new_title = g_strdup_printf ("%s (%d)",
+                                   is_today ? _("Today") : _("Scheduled"),
+                                   counter);
+    }
+
+  gtk_container_child_set (GTK_CONTAINER (priv->stack),
+                           container_child,
+                           "title", new_title,
+                           NULL);
+
+  g_list_free (tasks);
+  g_free (new_title);
+}
+
+
 static void
 gtd_window__change_storage_action (GSimpleAction *simple,
                                    GVariant      *parameter,
@@ -130,90 +163,6 @@ gtd_window__change_storage_action (GSimpleAction *simple,
   gtk_dialog_run (GTK_DIALOG (priv->storage_dialog));
 }
 
-static void
-gtd_window__notification_close_button_clicked (GtkButton *button,
-                                               gpointer   user_data)
-{
-  GtdWindowPrivate *priv;
-  NotificationData *data;
-
-  g_return_if_fail (GTD_IS_WINDOW (user_data));
-
-  priv = GTD_WINDOW (user_data)->priv;
-  data = g_queue_peek_head (priv->notification_queue);
-
-  /* Cancel any previouly set timeouts */
-  if (priv->notification_delay_id > 0)
-    {
-      g_source_remove (priv->notification_delay_id);
-      priv->notification_delay_id = 0;
-    }
-
-  gtd_window__execute_notification_data (data);
-}
-
-static void
-gtd_window_consume_notification (GtdWindow *window)
-{
-  GtdWindowPrivate *priv;
-  NotificationData *data;
-
-  g_return_if_fail (GTD_IS_WINDOW (window));
-
-  priv = window->priv;
-
-  if (g_queue_is_empty (priv->notification_queue))
-    {
-      gtk_revealer_set_reveal_child (priv->notification_revealer, FALSE);
-      priv->consuming_notifications = FALSE;
-      return;
-    }
-
-  priv->consuming_notifications = TRUE;
-
-  /* Keep the current operation until it is finished or canceled */
-  data = g_queue_peek_head (priv->notification_queue);
-
-  gtk_button_set_label (priv->notification_action_button, data->label ? data->label : "");
-  gtk_label_set_markup (priv->notification_label, data->text);
-  gtk_revealer_set_reveal_child (priv->notification_revealer, TRUE);
-  gtk_widget_set_visible (GTK_WIDGET (priv->notification_action_button), data->label != NULL);
-
-  /* spinner */
-  gtk_widget_set_visible (GTK_WIDGET (priv->notification_spinner), data->show_spinner);
-  g_object_set (priv->notification_spinner, "active", data->show_spinner, NULL);
-
-  /* If there's a delay set, execute the action */
-  if (data->delay)
-    {
-      priv->notification_delay_id = g_timeout_add (data->delay,
-                                                   (GSourceFunc) gtd_window__execute_notification_data,
-                                                   data);
-    }
-}
-
-static gboolean
-gtd_window__execute_notification_data (NotificationData *data)
-{
-  GtdWindowPrivate *priv = data->window->priv;
-  gboolean retval = G_SOURCE_REMOVE;
-
-  if (data->primary_action)
-    retval = data->primary_action (data->data);
-
-  /* Remove the current notification from queue */
-  g_queue_pop_head (priv->notification_queue);
-
-  /* Continue consuming notifications */
-  gtd_window_consume_notification (data->window);
-
-  notification_data_free (data);
-
-  priv->notification_delay_id = 0;
-
-  return retval;
-}
-
 static void
 gtd_window__list_color_set (GtkColorChooser *button,
                             gpointer         user_data)
@@ -223,9 +172,9 @@ gtd_window__list_color_set (GtkColorChooser *button,
   GdkRGBA new_color;
 
   g_return_if_fail (GTD_IS_WINDOW (user_data));
-  g_return_if_fail (gtd_list_view_get_task_list (priv->list_view));
+  g_return_if_fail (gtd_task_list_view_get_task_list (priv->list_view));
 
-  list = gtd_list_view_get_task_list (priv->list_view);
+  list = gtd_task_list_view_get_task_list (priv->list_view);
 
   g_debug ("%s: %s: %s",
            G_STRFUNC,
@@ -238,8 +187,28 @@ gtd_window__list_color_set (GtkColorChooser *button,
   gtd_manager_save_task_list (priv->manager, list);
 }
 
+static gboolean
+gtd_window__on_key_press_event (GtkWidget    *widget,
+                                GdkEvent     *event,
+                                GtkSearchBar *bar)
+{
+  GtdWindowPrivate *priv;
+
+  g_return_val_if_fail (GTD_IS_WINDOW (widget), TRUE);
+
+  priv = GTD_WINDOW (widget)->priv;
+
+  if (g_strcmp0 (gtk_stack_get_visible_child_name (GTK_STACK (priv->main_stack)), "overview") == 0 &&
+      g_strcmp0 (gtk_stack_get_visible_child_name (GTK_STACK (priv->stack)), "lists") == 0)
+    {
+      return gtk_search_bar_handle_event (bar, event);
+    }
+
+  return FALSE;
+}
+
 static gint
-gtd_window__listbox_sort_func (GtdTaskListItem *a,
+gtd_window__flowbox_sort_func (GtdTaskListItem *a,
                                GtdTaskListItem *b,
                                gpointer         user_data)
 {
@@ -258,29 +227,60 @@ gtd_window__listbox_sort_func (GtdTaskListItem *a,
   return g_strcmp0 (gtd_task_list_get_name (l1), gtd_task_list_get_name (l2));
 }
 
+static gboolean
+gtd_window__flowbox_filter_func (GtdTaskListItem *item,
+                                 GtdWindow       *window)
+{
+  GtdWindowPrivate *priv;
+  GtdTaskList *list;
+  gboolean return_value;
+  gchar *search_folded;
+  gchar *list_name_folded;
+  gchar *haystack;
+
+  g_return_val_if_fail (GTD_IS_WINDOW (window), FALSE);
+
+  priv = window->priv;
+  list = gtd_task_list_item_get_list (item);
+  haystack = NULL;
+  search_folded = g_utf8_casefold (gtk_entry_get_text (GTK_ENTRY (priv->search_entry)), -1);
+  list_name_folded = g_utf8_casefold (gtd_task_list_get_name (list), -1);
+
+  if (!search_folded || search_folded[0] == '\0')
+    {
+      return_value = TRUE;
+      goto out;
+    }
+
+  haystack = g_strstr_len (list_name_folded,
+                           -1,
+                           search_folded);
+
+  return_value = (haystack != NULL);
+
+out:
+  g_free (search_folded);
+  g_free (list_name_folded);
+
+  return return_value;
+}
+
 static void
 gtd_window__manager_ready_changed (GObject    *source,
                                    GParamSpec *spec,
                                    gpointer    user_data)
 {
   GtdWindowPrivate *priv = GTD_WINDOW (user_data)->priv;
+
   g_return_if_fail (GTD_IS_WINDOW (user_data));
 
   if (gtd_object_get_ready (GTD_OBJECT (source)))
     {
-      gtd_window_cancel_notification (GTD_WINDOW (user_data), LOADING_LISTS_NOTIFICATION_ID);
+      gtd_notification_widget_cancel (priv->notification_widget, priv->loading_notification);
     }
   else
     {
-      gtd_window_notify (GTD_WINDOW (user_data),
-                         0,
-                         LOADING_LISTS_NOTIFICATION_ID,
-                         _("Loading your task lists…"),
-                         NULL,
-                         NULL,
-                         NULL,
-                         TRUE,
-                         NULL);
+      gtd_notification_widget_notify (priv->notification_widget, priv->loading_notification);
     }
 }
 
@@ -294,6 +294,7 @@ gtd_window__back_button_clicked (GtkButton *button,
 
   gtk_stack_set_visible_child_name (priv->main_stack, "overview");
   gtk_header_bar_set_custom_title (priv->headerbar, GTK_WIDGET (priv->stack_switcher));
+  gtk_header_bar_set_title (priv->headerbar, _("To Do"));
   gtk_widget_hide (GTK_WIDGET (priv->back_button));
   gtk_widget_hide (GTK_WIDGET (priv->color_button));
 }
@@ -323,8 +324,9 @@ gtd_window__list_selected (GtkFlowBox      *flowbox,
   gtk_header_bar_set_title (priv->headerbar, gtd_task_list_get_name (list));
   gtk_header_bar_set_subtitle (priv->headerbar, gtd_task_list_get_origin (list));
   gtk_header_bar_set_custom_title (priv->headerbar, NULL);
-  gtd_list_view_set_task_list (priv->list_view, list);
-  gtd_list_view_set_show_completed (priv->list_view, FALSE);
+  gtk_search_bar_set_search_mode (priv->search_bar, FALSE);
+  gtd_task_list_view_set_task_list (priv->list_view, list);
+  gtd_task_list_view_set_show_completed (priv->list_view, FALSE);
   gtk_widget_show (GTK_WIDGET (priv->back_button));
   gtk_widget_show (GTK_WIDGET (priv->color_button));
 
@@ -359,10 +361,15 @@ gtd_window_constructed (GObject *object)
   G_OBJECT_CLASS (gtd_window_parent_class)->constructed (object);
 
   gtk_flow_box_set_sort_func (priv->lists_flowbox,
-                              (GtkFlowBoxSortFunc) gtd_window__listbox_sort_func,
+                              (GtkFlowBoxSortFunc) gtd_window__flowbox_sort_func,
                               NULL,
                               NULL);
 
+  gtk_flow_box_set_filter_func (priv->lists_flowbox,
+                                (GtkFlowBoxFilterFunc) gtd_window__flowbox_filter_func,
+                                object,
+                                NULL);
+
   g_object_bind_property (object,
                           "manager",
                           priv->new_list_popover,
@@ -424,9 +431,9 @@ gtd_window_set_property (GObject      *object,
     case PROP_MANAGER:
       self->priv->manager = g_value_get_object (value);
 
-      gtd_list_view_set_manager (self->priv->list_view, self->priv->manager);
-      gtd_list_view_set_manager (self->priv->today_list_view, self->priv->manager);
-      gtd_list_view_set_manager (self->priv->scheduled_list_view, self->priv->manager);
+      gtd_task_list_view_set_manager (self->priv->list_view, self->priv->manager);
+      gtd_task_list_view_set_manager (self->priv->today_list_view, self->priv->manager);
+      gtd_task_list_view_set_manager (self->priv->scheduled_list_view, self->priv->manager);
 
       g_signal_connect (self->priv->manager,
                         "notify::ready",
@@ -450,8 +457,33 @@ gtd_window_set_property (GObject      *object,
       g_list_free (lists);
 
       /* Setup 'Today' and 'Scheduled' lists */
-      gtd_list_view_set_task_list (self->priv->today_list_view, gtd_manager_get_today_list (self->priv->manager));
-      gtd_list_view_set_task_list (self->priv->scheduled_list_view, gtd_manager_get_scheduled_list (self->priv->manager));
+      gtd_task_list_view_set_task_list (self->priv->today_list_view, gtd_manager_get_today_list (self->priv->manager));
+      gtd_task_list_view_set_task_list (self->priv->scheduled_list_view, gtd_manager_get_scheduled_list (self->priv->manager));
+
+      g_signal_connect (gtd_manager_get_today_list (self->priv->manager),
+                        "task-added",
+                        G_CALLBACK (gtd_window_update_list_counters),
+                        self);
+      g_signal_connect (gtd_manager_get_today_list (self->priv->manager),
+                        "task-updated",
+                        G_CALLBACK (gtd_window_update_list_counters),
+                        self);
+      g_signal_connect (gtd_manager_get_today_list (self->priv->manager),
+                        "task-removed",
+                        G_CALLBACK (gtd_window_update_list_counters),
+                        self);
+      g_signal_connect (gtd_manager_get_scheduled_list (self->priv->manager),
+                        "task-added",
+                        G_CALLBACK (gtd_window_update_list_counters),
+                        self);
+      g_signal_connect (gtd_manager_get_scheduled_list (self->priv->manager),
+                        "task-updated",
+                        G_CALLBACK (gtd_window_update_list_counters),
+                        self);
+      g_signal_connect (gtd_manager_get_scheduled_list (self->priv->manager),
+                        "task-removed",
+                        G_CALLBACK (gtd_window_update_list_counters),
+                        self);
 
       g_object_notify (object, "manager");
       break;
@@ -495,19 +527,21 @@ gtd_window_class_init (GtdWindowClass *klass)
   gtk_widget_class_bind_template_child_private (widget_class, GtdWindow, list_view);
   gtk_widget_class_bind_template_child_private (widget_class, GtdWindow, main_stack);
   gtk_widget_class_bind_template_child_private (widget_class, GtdWindow, new_list_popover);
-  gtk_widget_class_bind_template_child_private (widget_class, GtdWindow, notification_action_button);
-  gtk_widget_class_bind_template_child_private (widget_class, GtdWindow, notification_label);
-  gtk_widget_class_bind_template_child_private (widget_class, GtdWindow, notification_revealer);
-  gtk_widget_class_bind_template_child_private (widget_class, GtdWindow, notification_spinner);
+  gtk_widget_class_bind_template_child_private (widget_class, GtdWindow, notification_widget);
+  gtk_widget_class_bind_template_child_private (widget_class, GtdWindow, scheduled_list_view);
+  gtk_widget_class_bind_template_child_private (widget_class, GtdWindow, stack);
+  gtk_widget_class_bind_template_child_private (widget_class, GtdWindow, search_bar);
+  gtk_widget_class_bind_template_child_private (widget_class, GtdWindow, search_button);
+  gtk_widget_class_bind_template_child_private (widget_class, GtdWindow, search_entry);
   gtk_widget_class_bind_template_child_private (widget_class, GtdWindow, stack_switcher);
   gtk_widget_class_bind_template_child_private (widget_class, GtdWindow, storage_dialog);
-  gtk_widget_class_bind_template_child_private (widget_class, GtdWindow, scheduled_list_view);
   gtk_widget_class_bind_template_child_private (widget_class, GtdWindow, today_list_view);
 
   gtk_widget_class_bind_template_callback (widget_class, gtd_window__back_button_clicked);
   gtk_widget_class_bind_template_callback (widget_class, gtd_window__list_color_set);
   gtk_widget_class_bind_template_callback (widget_class, gtd_window__list_selected);
-  gtk_widget_class_bind_template_callback (widget_class, gtd_window__notification_close_button_clicked);
+  gtk_widget_class_bind_template_callback (widget_class, gtd_window__on_key_press_event);
+  gtk_widget_class_bind_template_callback (widget_class, gtd_window__stack_visible_child_cb);
 }
 
 static void
@@ -515,7 +549,8 @@ gtd_window_init (GtdWindow *self)
 {
   self->priv = gtd_window_get_instance_private (self);
 
-  self->priv->notification_queue = g_queue_new ();
+  self->priv->loading_notification = gtd_notification_new (_("Loading your task lists…"), 0);
+  gtd_object_set_ready (GTD_OBJECT (self->priv->loading_notification), FALSE);
 
   /* add actions */
   g_action_map_add_action_entries (G_ACTION_MAP (self),
@@ -545,111 +580,43 @@ gtd_window_get_manager (GtdWindow *window)
 /**
  * gtd_window_notify:
  * @window: a #GtdWindow
- * @visible_time: time which the notification will be visible, 0 to show permanently
- * @text: text of the notification
- * @button_label: text to show on the notification button, %NULL to hide it
- * @primary_action: function to call when close button is clicked
- * @secondary_action: function to call when the alternative button is clicked
- * @show_spinner: whether show the spinner or not
- * @user_data: custom data
+ * @notification: a #GtdNotification
  *
  * Shows a notification on the top of the main window.
  *
  * Returns:
  */
 void
-gtd_window_notify (GtdWindow   *window,
-                   gint         visible_time,
-                   const gchar *id,
-                   const gchar *text,
-                   const gchar *button_label,
-                   GSourceFunc  primary_action,
-                   GSourceFunc  secondary_action,
-                   gboolean     show_spinner,
-                   gpointer     user_data)
+gtd_window_notify (GtdWindow       *window,
+                   GtdNotification *notification)
 {
   GtdWindowPrivate *priv;
-  NotificationData *data;
 
   g_return_if_fail (GTD_IS_WINDOW (window));
 
   priv = window->priv;
 
-  data = g_new0 (NotificationData, 1);
-  data->window = window;
-  data->delay = visible_time;
-  data->id = g_strdup (id);
-  data->text = g_strdup (text);
-  data->label = g_strdup (button_label);
-  data->primary_action = primary_action;
-  data->secondary_action = secondary_action;
-  data->show_spinner = show_spinner;
-  data->data = user_data;
-
-  g_queue_push_tail (priv->notification_queue, data);
-
-  /* If we're not consuming notifications, start it now */
-  if (!priv->consuming_notifications)
-    gtd_window_consume_notification (window);
+  gtd_notification_widget_notify (priv->notification_widget, notification);
 }
 
 /**
  * gtd_window_cancel_notification:
  * @window: a #GtdManager
- * @id: id of the given notification
+ * @notification: a #GtdNotification
  *
- * Cancels the notification with @id as id.
+ * Cancels @notification.
  *
  * Returns:
  */
 void
-gtd_window_cancel_notification (GtdWindow   *window,
-                                const gchar *id)
+gtd_window_cancel_notification (GtdWindow       *window,
+                                GtdNotification *notification)
 {
   GtdWindowPrivate *priv;
-  NotificationData *data;
-  GList *head;
-  GList *l;
-  gint index;
 
   g_return_if_fail (GTD_IS_WINDOW (window));
-  g_return_if_fail (id != NULL);
 
-  data = NULL;
   priv = window->priv;
-  index = 0;
 
-  /* Search for a notification with the given id */
-  head = priv->notification_queue->head;
-
-  for (l = head; l != NULL; l = l->next)
-    {
-      NotificationData *tmp = l->data;
-
-      if (tmp && g_strcmp0 (tmp->id, id) == 0)
-        {
-          data = l->data;
-          break;
-        }
-
-      index++;
-    }
-
-  g_queue_remove (priv->notification_queue, data);
-
-  /* If we're removing the current head, continue consuming the queue */
-  if (index == 0)
-    {
-      /* Remove any remaining timeouts */
-      if (priv->notification_delay_id > 0)
-        {
-          g_source_remove (priv->notification_delay_id);
-          priv->notification_delay_id = 0;
-        }
-
-      gtd_window_consume_notification (window);
-    }
-
-  if (data)
-    notification_data_free (data);
+  gtd_notification_widget_cancel (priv->notification_widget, notification);
 }
